# Standard library imports
from os.path import join as pjoin, dirname
from distutils.dep_util import newer_group
from distutils.errors import DistutilsError

from numpy.distutils.misc_util import appendpath
from numpy.distutils import log


def generate_a_pyrex_source(self, base, ext_name, source, extension):
    """ Monkey patch for numpy build_src.build_src method

    Uses Cython instead of Pyrex.

    Assumes Cython is present
    """
    if self.inplace:
        target_dir = dirname(base)
    else:
        target_dir = appendpath(self.build_src, dirname(base))
    target_file = pjoin(target_dir, ext_name + '.c')
    depends = [source] + extension.depends
    if self.force or newer_group(depends, target_file, 'newer'):
        import Cython.Compiler.Main
        log.info("cythonc:> %s" % (target_file))
        self.mkpath(target_dir)
        options = Cython.Compiler.Main.CompilationOptions(
            defaults=Cython.Compiler.Main.default_options,
            include_path=extension.include_dirs,
            output_file=target_file)
        cython_result = Cython.Compiler.Main.compile(source,
                                                   options=options)
        if cython_result.num_errors != 0:
            raise DistutilsError("%d errors while compiling %r with Cython" \
                  % (cython_result.num_errors, source))
    return target_file


from numpy.distutils.command import build_src
build_src.build_src.generate_a_pyrex_source = generate_a_pyrex_source



import sys
import os
#import setuptools

linkArgs = []
if sys.platform == 'win32':
    linkArgs = ['-static-libgcc']
    
    #if sys.version_info[0] == 3:
    #    linkArgs.append('-l python3.lib')


#windows VC++ has really shocking c standard support so we need to include
#custom stdint.h and intypes.h files from https://code.google.com/archive/p/msinttypes
#print os.environ.get('CC', 'foo')
if False: #sys.platform == 'win32' and not os.environ.get('CC', '') == 'mingw':
    extra_include_dirs = ['win_incl']
else:
    extra_include_dirs = []

#import cython_numpy_monkey
#import setuptools

#from Cython.Distutils import build_ext

from Cython.Build import cythonize


#############
# Monkey-patch distutils to not link MSVCR90
import numpy.distutils.mingw32ccompiler
from distutils.unixccompiler import UnixCCompiler
#numpy.distutils.mingw32ccompiler.msvc_runtime_library = lambda : None
numpy.distutils.mingw32ccompiler.build_msvcr_library = lambda debug=False : False

def link(self,
             target_desc,
             objects,
             output_filename,
             output_dir,
             libraries,
             library_dirs,
             runtime_library_dirs,
             export_symbols = None,
             debug=0,
             extra_preargs=None,
             extra_postargs=None,
             build_temp=None,
             target_lang=None):
        # Include the appropiate MSVC runtime library if Python was built
        # with MSVC >= 7.0 (MinGW standard is msvcrt)
        #runtime_library = msvc_runtime_library()
        #if runtime_library:
        #    if not libraries:
        #        libraries = []
        #    libraries.append(runtime_library)
        args = (self,
                target_desc,
                objects,
                output_filename,
                output_dir,
                libraries,
                library_dirs,
                runtime_library_dirs,
                None, #export_symbols, we do this in our def-file
                debug,
                extra_preargs,
                extra_postargs,
                build_temp,
                target_lang)
        if self.gcc_version < "3.0.0":
            import distutils
            func = distutils.cygwinccompiler.CygwinCCompiler.link
        else:
            func = UnixCCompiler.link
        func(*args[:func.__code__.co_argcount])
        return

numpy.distutils.mingw32ccompiler.Mingw32CCompiler.link = link



####
# Python3 windows patch: (here for now to document, should ideally create a better workaround).
# NB - seems to also be needed for recent python 2 (is this a numpy version thing instead?)
#
# modify numpy.distutils.exec_command._exec_command to add line: command[0] = find_executable(command[0])
#

import numpy.distutils.exec_command

def _monkey_patch_exec_command(fcn):
    def _exec_command(command, use_shell=None, use_tee=None, **env):
        command[0] = numpy.distutils.exec_command.find_executable(command[0])
        return fcn(command, use_shell=use_shell, use_tee=use_tee, **env)
        
    return _exec_command
    
if sys.platform == 'win32':
    numpy.distutils.exec_command._exec_command = _monkey_patch_exec_command(numpy.distutils.exec_command._exec_command)

### End exec_command patching
import subprocess
from numpy.distutils.exec_command import find_executable

#objdump is a .bat rather than a .exe in an anaconda build environment
def _monkey_patch_dump_table(dll):
    objdump = find_executable("objdump")
    st = subprocess.Popen([objdump, "-p", dll], stdout=subprocess.PIPE)
    return st.stdout.readlines()

numpy.distutils.mingw32ccompiler.dump_table = _monkey_patch_dump_table

# ditto for dlltool
def _build_import_library_amd64():
    out_exists, out_file = numpy.distutils.mingw32ccompiler._check_for_import_lib()
    if out_exists:
        numpy.distutils.mingw32ccompiler.log.debug('Skip building import library: "%s" exists', out_file)
        return

    # get the runtime dll for which we are building import library
    dll_file = numpy.distutils.mingw32ccompiler.find_python_dll()
    numpy.distutils.mingw32ccompiler.log.info('Building import library (arch=AMD64): "%s" (from %s)' %
             (out_file, dll_file))

    # generate symbol list from this library
    def_name = "python%d%d.def" % tuple(sys.version_info[:2])
    def_file = os.path.join(sys.prefix, 'libs', def_name)
    numpy.distutils.mingw32ccompiler.generate_def(dll_file, def_file)

    # generate import library from this symbol list
    cmd = [find_executable("dlltool"), '-d', def_file, '-l', out_file]
    subprocess.Popen(cmd)

numpy.distutils.mingw32ccompiler._build_import_library_amd64 = _build_import_library_amd64

# End monkey patching
#####################

#pre-compile cython (copied & adapted from skimage._build)
def cython(pyx_files, working_path=''):
    """Use Cython to convert the given files to C.
    Parameters
    ----------
    pyx_files : list of str
        The input .pyx files.
    """
    # Do not build cython files if target is clean
    if len(sys.argv) >= 2 and sys.argv[1] == 'clean':
        return

    try:

        from Cython.Build import cythonize
    except ImportError:
        # If cython is not found, the build will make use of
        # the distributed .c or .cpp files if present
       

        print("Cython not found; falling back to pre-built bcl.c")
    else:
        from multiprocessing import cpu_count
        pyx_files = [os.path.join(working_path, f) for f in pyx_files]

        # Cython doesn't automatically choose a number of threads > 1
        # https://github.com/cython/cython/blob/a0bbb940c847dfe92cac446c8784c34c28c92836/Cython/Build/Dependencies.py#L923-L925
        cythonize(pyx_files, nthreads=cpu_count(),
                  compiler_directives={'language_level': sys.version_info.major})

def configuration(parent_package = '', top_path = None):
    from numpy.distutils.core import Extension
    from numpy.distutils.misc_util import Configuration, get_numpy_include_dirs

    if not parent_package == '':
        name = '.'.join([parent_package, 'pymecompress', 'bcl'])
    else:
        name = '.'.join(['pymecompress', 'bcl'])
    
    #cythonize pyx files
    cython(['bcl.pyx',], working_path=os.path.abspath(os.path.dirname(__file__)))

    config = Configuration('pymecompress', parent_package, top_path)
    config.get_version()
    
    config.add_extension(name='bcl',
                    #sources=[os.path.join(cur_dir, 'bcl.pyx'), os.path.join(cur_dir, 'bcl/huffman.c'), os.path.join(cur_dir, 'quantize.c')],
                    sources=['bcl.c', 'bcl/huffman.c','quantize.dispatch.c', 'quantize.c'],
                    include_dirs=['bcl',] + get_numpy_include_dirs() + extra_include_dirs,
                    extra_compile_args=['-O3', '-fno-exceptions', '-ffast-math',],
                    extra_link_args=linkArgs)

    

    return config


# import distutils
# class build_ext_subclass(distutils.command.build_ext.build_ext):
#     def build_extensions(self):

#         original__compile = self.compiler._compile
#         def new__compile(obj, src, ext, cc_args, extra_postargs, pp_opts):
#             if src.endswith("_AVX.c"):
#                 extra_postargs = extra_postargs + ['-mavx',]
#             return original__compile(obj, src, ext, cc_args, extra_postargs, pp_opts)
#         self.compiler._compile = new__compile
#         try:
#             distutils.command.build_ext.build_ext.build_extensions(self)
#         finally:
#             del self.compiler._compile

if __name__ == '__main__':
    from numpy.distutils.core import setup
    
    setup(description = 'Compression for photon-noise limited images which keeps losses within the Poisson noise envelope',
          author = 'David Baddeley',
          author_email = 'david.baddeley@yale.edu',
          url = 'https://github.com/python-microscopy/pymecompress',
          long_description = 'Compression for photon-noise limited images which keeps losses within the Poisson noise envelope',
          license = "BSD",
          install_requires=['numpy'],
          classifiers=[
              'Development Status :: 3 - Alpha',
              # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
              'License :: OSI Approved :: BSD License', # Again, pick a license
              'Programming Language :: Python :: 2.7', #Specify which pyhton versions that you want to support
              'Programming Language :: Python :: 3.6',
              'Programming Language :: Python :: 3.7',
          ],
          **configuration(top_path='').todict()
         )




#!/usr/bin/python

##################
# setup.py
#
# Copyright David Baddeley, 2009-2020
# d.baddeley@auckland.ac.nz
#
#
##################
from numpy.distutils.command import build_ext, build_src
import os

#Monkey-patch to generate directory for dispatch headers. NOTE - this is fixed in current numpy head (3/4/2022), but not in latest
#release (1.22.x)
class bsrc(build_src.build_src):
    def finalize_options(self):
        super().finalize_options()
        print('build_src: ', getattr(self, 'build_src', None))

        os.makedirs(os.path.join(getattr(self, 'build_src'), 'pymecompress'), exist_ok=True)
        #print('build_temp: ', getattr(self, 'build_temp', None))

def configuration(parent_package='', top_path=None):
    if os.path.exists('MANIFEST'):
        os.remove('MANIFEST')
    
    from numpy.distutils.misc_util import Configuration
    config = Configuration(None, parent_package, top_path)
    
    config.set_options(
        ignore_setup_xxx_py=True,
        assume_default_configuration=True,
        delegate_options_to_subpackages=True,
        #quiet=True,
    )
    
    config.add_subpackage('pymecompress')
    config.get_version('pymecompress/version.py')
    return config


if __name__ == '__main__':
    import setuptools
    import os
    from distutils.command.sdist import sdist
    from numpy.distutils.core import setup


    with open('README.md', 'r') as f:
        long_description = f.read()

    setup(name='pymecompress',
        description='Compression for photon-noise limited images which keeps losses within the Poisson noise envelope',
        author='David Baddeley',
        author_email='david.baddeley@yale.edu',
        url='https://github.com/python-microscopy/pymecompress',
        long_description=long_description,
        long_description_content_type="text/markdown",
        license="BSD",
        install_requires=['numpy'],
        classifiers=[
            'Development Status :: 3 - Alpha',
            # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
            'License :: OSI Approved :: BSD License', # Again, pick a license
            #'Programming Language :: Python :: 2.7', #Specify which python versions that you want to support
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
        ],
        cmdclass={'sdist': sdist, 'build_src':bsrc},
        configuration=configuration)
    
    



import numpy as np
#import six
from cython.view cimport array as cvarray
cimport cython
from libc.stdint cimport uint16_t, uint8_t, uint32_t
from cpython cimport PyObject_CheckBuffer, PyObject_GetBuffer, PyBuffer_Release, Py_buffer, PyObject, PyBUF_SIMPLE, PyBUF_C_CONTIGUOUS, PyBUF_F_CONTIGUOUS, PyBuffer_IsContiguous

cdef extern from "bcl/huffman.h":
    int Huffman_Compress( unsigned char *inp, unsigned char *out, unsigned int insize ) nogil
    int Huffman_Compress_( unsigned char *inp, unsigned char *out, unsigned int insize ) nogil
    void Huffman_Uncompress( unsigned char *inp, unsigned char *out, unsigned int insize, unsigned int outsize ) nogil

cdef extern from "quantize.h":
    void quantize_u16(uint16_t *data, uint8_t * out, int size, float offset, float scale) nogil
    #void quantize_u16_avx( uint16_t * data, uint8_t * out, int size, float offset, float scale) nogil
    
@cython.boundscheck(False)
def HuffmanCompress(data):
    cdef Py_buffer view
    
    #print('HuffmanCompress')
    
    if data.flags['C_CONTIGUOUS']:
        PyObject_GetBuffer(data, &view, PyBUF_C_CONTIGUOUS)
        #print('saving c contiguous')
        #raise RuntimeError('C contig')
    elif data.flags['F_CONTIGUOUS']:
        PyObject_GetBuffer(data, &view, PyBUF_F_CONTIGUOUS)
        #print('saving fortran contiguous')
        #raise RuntimeError('F contig')
    else:
        raise RuntimeError('Input data should be contiguous')
    
    cdef int dsize = view.len
    
    out = np.zeros(int(dsize*1.01 + 320),'uint8')
    cdef unsigned char [:] ov = out
    
    with nogil:
        
        nb = Huffman_Compress(<uint8_t *>view.buf, &ov[0], dsize)
        
    PyBuffer_Release(&view)
    return out[:nb]

@cython.boundscheck(False)
def huffman_compress_buffer(data):
    cdef Py_buffer buffer
    PyObject_GetBuffer(data, &buffer, PyBUF_C_CONTIGUOUS)
    #assert(PyBuffer_IsContiguous(buffer, 'C'))
    cdef int nb
    cdef int dsize = buffer.len
    cdef int orig_size = int(buffer.len/buffer.itemsize)
    
    out = np.zeros(int(dsize*1.01 + 320),'uint8')
    cdef unsigned char [:] ov = out
    
    with nogil:
        
        nb = Huffman_Compress(<uint8_t *>buffer.buf, &ov[0], dsize)
        
    PyBuffer_Release(&buffer)
    # store length in last 4 bytes
    (<uint32_t *>(&ov[nb]))[0] = orig_size
    return out[:(nb + 4)]

@cython.boundscheck(False)
def huffman_compress_quant_buffer(data, float offset, float scale):
    #assert(PyBuffer_IsContiguous(buffer, 'C'))
    cdef Py_buffer buffer
    PyObject_GetBuffer(data, &buffer, PyBUF_C_CONTIGUOUS)
    
    cdef int nb
    cdef int dsize = buffer.len
    cdef int orig_size = int(buffer.len/buffer.itemsize)
    
    out = np.zeros(int(dsize*1.01 + 320),'uint8')
    quant = np.zeros(dsize, 'uint8')
    cdef unsigned char [:] ov = out
    cdef unsigned char [:] qv = quant
    
    with nogil:
        quantize_u16(<uint16_t *>buffer.buf, &qv[0], dsize, offset, scale)
        nb = Huffman_Compress(&qv[0], &ov[0], dsize)
        
    PyBuffer_Release(&buffer)
    # store length in last 4 bytes
    (<uint32_t *>(&ov[nb]))[0] = orig_size
    return out[:(nb + 4)]

@cython.boundscheck(False)
def HuffmanCompressQuant(data, float offset, float scale):
    cdef Py_buffer view
    
    if not data.dtype == 'u2':
        raise RuntimeError('Expected unsigned short input data')
    
    if data.flags['C_CONTIGUOUS']:
        PyObject_GetBuffer(data, &view, PyBUF_C_CONTIGUOUS)
    elif data.flags['F_CONTIGUOUS']:
        PyObject_GetBuffer(data, &view, PyBUF_F_CONTIGUOUS)
    else:
        raise RuntimeError('Input data should be contiguous')
    
    cdef int dsize = data.size
    
    out = np.zeros(int(dsize*1.01 + 320),'uint8')
    quant = np.zeros(dsize, 'uint8')
    cdef unsigned char [:] ov = out
    cdef unsigned char [:] qv = quant
    
    with nogil:
        quantize_u16(<uint16_t *>view.buf, &qv[0], dsize, offset, scale)
        nb = Huffman_Compress(&qv[0], &ov[0], dsize)
        
    PyBuffer_Release(&view)
    return out[:nb]

@cython.boundscheck(False)    
def HuffmanCompressOrig(unsigned char[:] data):
    out = np.zeros(int(data.shape[0]*1.01 + 320),'uint8')
    cdef unsigned char [:] ov = out
    cdef int dsize = data.shape[0]
    with nogil:
        
        nb = Huffman_Compress_(&data[0], &ov[0], dsize)
    return out[:nb]

@cython.boundscheck(False)   
def HuffmanDecompress(unsigned char[:] data, unsigned int outsize):
    out = np.zeros(outsize,'uint8')
    cdef unsigned char [:] ov = out
    cdef int insize = data.shape[0]
    #cdef int outsize = outsize
    with nogil:
        
        Huffman_Uncompress(&data[0], &ov[0], insize, outsize)
    return out
    
@cython.boundscheck(False)
def huffman_decompress_buffer(data,  out):
    cdef Py_buffer buffer
    PyObject_GetBuffer(data, &buffer, PyBUF_C_CONTIGUOUS)
    cdef Py_buffer outb
    #assert(PyBuffer_IsContiguous(data, 'C'))
    cdef int outlen = (<uint32_t *>(&(<uint8_t *>buffer.buf)[buffer.len-4]))[0]
    #print('outlen:', outlen)
    
    if out is None:
        out = np.zeros(outlen, 'uint8')
    
    PyObject_GetBuffer(out, &outb, PyBUF_C_CONTIGUOUS)
    assert(outb.len == outlen)
    
    with nogil:
        Huffman_Uncompress(<uint8_t *>buffer.buf, <uint8_t *>outb.buf, buffer.len-4, outb.len)
    
    PyBuffer_Release(&buffer)
    PyBuffer_Release(&outb)
    
    return out
       


"""
numcodecs compatible compression and quantization codecs.
"""
from . import bcl
import numcodecs
from numcodecs.abc import Codec


class Huffman(Codec):
    codec_id='pymecompress-huffman'
    
    def encode(self, buf):
        return bcl.huffman_compress_buffer(buf)
    
    def decode(self, buf, out=None):
        return bcl.huffman_decompress_buffer(buf, out)
        
    def get_config(self):
        return {'id': self.codec_id}
    
    @classmethod
    def from_config(cls, config):
        return cls()

numcodecs.register_codec(Huffman)

class HuffmanQuant16(Codec):
    codec_id = 'pymecompress-quant16'
    
    def __init__(self, offset=0, scale=1):
        self._offset = offset
        self._scale = scale
    
    def encode(self, buf):
        return bcl.huffman_compress_quant_buffer(buf, self._offset, self._scale)
    
    def decode(self, buf, out=None):
        ret = bcl.huffman_decompress_buffer(buf, None).astype('uint16')
        
        ret = (ret*ret)/self._scale + self._offset
        
        if out is None:
            out = ret
        else:
           out[:] = ret
        
        return out.astype('uint16')
    
    def get_config(self):
        return {'id': self.codec_id,
                'offset': self._offset, 'scale' : self._scale}
    
    @classmethod
    def from_config(cls, config):
        return cls(offset=config.get('offset', 0), scale=config.get('scale', 1))

numcodecs.register_codec(HuffmanQuant16)
